import logging
import os
import sys
from datetime import datetime

import pytz
import urllib3
import yaml

from jf_ingest.config import GitAuthConfig, GitConfig
from jf_ingest.jf_git.adapters import load_and_push_git_to_s3
from jf_ingest.jf_jira import IngestionConfig, load_and_push_jira_to_s3
from jf_ingest.jf_jira.auth import JiraDownloadConfig
from jf_ingest.validation import validate_git, validate_jira

logger = logging.getLogger(__name__)

# Start of Epoch Time
_default_datetime_str = '1970-01-01'


def setup_harness_logging(logging_level: int):
    """Helper function to setting up logging in the harness"""
    logging.basicConfig(
        level=logging_level,
        format=(
            "%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s"
            if logging_level == logging.DEBUG
            else "%(message)s"
        ),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


def _process_jira_config(general_config_data: dict) -> JiraDownloadConfig:
    jira_config_data = general_config_data["jira_config"]
    # Translate Datetimes first
    jira_config_data["work_logs_pull_from"] = datetime.strptime(
        jira_config_data.get("work_logs_pull_from", _default_datetime_str), "%Y-%m-%d"
    )
    jira_config_data["earliest_issue_dt"] = datetime.strptime(
        jira_config_data.get("earliest_issue_dt", _default_datetime_str), "%Y-%m-%d"
    )
    # Generate object in memory
    jira_config = JiraDownloadConfig(**jira_config_data)
    jira_config.url = os.getenv("JIRA_URL")
    jira_config.user = os.getenv("JIRA_USERNAME")
    jira_config.password = os.getenv("JIRA_PASSWORD")
    return jira_config


def _process_git_configs(general_config_data: dict) -> list[GitConfig]:
    git_config_data = general_config_data["git_configs"][0]
    company_slug = git_config_data['company_slug']
    # Process 'default' time for repos and commits
    git_config_data["default_pull_from_for_commits_and_prs"] = datetime.strptime(
        git_config_data.get("default_pull_from_for_commits_and_prs", _default_datetime_str),
        "%Y-%m-%d",
    ).replace(tzinfo=pytz.UTC)

    def _process_pr_or_commits_pull_from_values(commits_or_prs: str):
        # Translate pull_commits_since_for_repo_in_org
        for org_login in git_config_data.get(
            f'pull_{commits_or_prs}_since_for_repo_in_org', {}
        ).keys():
            for repo_id, date_str in git_config_data[
                f'pull_{commits_or_prs}_since_for_repo_in_org'
            ][org_login].items():
                git_config_data[f'pull_{commits_or_prs}_since_for_repo_in_org'][org_login][
                    repo_id
                ] = datetime.strptime(date_str, "%Y-%m-%d").replace(tzinfo=pytz.UTC)

    _process_pr_or_commits_pull_from_values('commits')
    _process_pr_or_commits_pull_from_values('prs')

    # Create separate auth config
    auth_config = GitAuthConfig(
        company_slug=company_slug,
        token=os.getenv('GITHUB_TOKEN'),
        base_url='',
        verify=False,
        session=None,
    )
    return [GitConfig(git_auth_config=auth_config, **git_config_data)]


if __name__ == "__main__":
    """
    NOTE: This is a work in progress developer debugging tool.
    it is currently run by using the following command:
       pdm run ingest_harness [--debug]
    and it requires you to have a creds.env and a config.yml file at
    the root of this project
    """
    debug_mode = "--debug" in sys.argv
    validate_mode = "--validate" in sys.argv
    run_git_harness = '--git' in sys.argv
    run_jira_harness = '--jira' in sys.argv

    if not run_git_harness and not run_jira_harness:
        # If neither the git or jira arg are supplied, run validation for both
        run_git_harness = True
        run_jira_harness = True

    # Get Config data for Ingestion Config
    with open("./config.yml") as yaml_file:
        yaml_data = yaml.safe_load(yaml_file)
        general_config_data = yaml_data["general"]
        ingest_config = IngestionConfig(
            timestamp=datetime.utcnow().strftime("%Y%m%d_%H%M%S"),
            jellyfish_api_token=os.getenv("JELLYFISH_API_TOKEN"),
            **general_config_data,
        )
        ingest_config.company_slug = (os.getenv("COMPANY_SLUG"),)
        ingest_config.local_file_path = f"{ingest_config.local_file_path}/{ingest_config.timestamp}"

        # Processes Jira Info
        if run_jira_harness:
            if 'jira_config' in general_config_data:
                ingest_config.jira_config = _process_jira_config(general_config_data)
            else:
                logger.warning(
                    f'Attempted to run Jira Validation but the Jira Config was not found in the config.yml file'
                )

        # Process Git Info
        if run_git_harness:
            if 'git_configs' in general_config_data:
                ingest_config.git_configs = _process_git_configs(general_config_data)
            else:
                logger.warning(
                    f'Attempting to run Git Validation but the Git Config was not found in the config.yml file'
                )

    setup_harness_logging(logging_level=logging.DEBUG if debug_mode else logging.INFO)

    if validate_mode:
        if run_jira_harness and 'jira_config' in general_config_data:
            validate_jira(ingest_config.jira_config)
        if run_git_harness and 'git_configs' in general_config_data:
            validate_git(ingest_config.git_configs)
    else:
        if run_jira_harness:
            load_and_push_jira_to_s3(ingest_config)
        if run_git_harness:
            load_and_push_git_to_s3(ingest_config)



import logging
from enum import Enum

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import IngestionConfig, IngestionType, IssueMetadata, JiraAuthMethod
from jf_ingest.constants import Constants
from jf_ingest.file_operations import IngestIOHelper, SubDirectory
from jf_ingest.jf_jira.auth import get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    download_all_issue_metadata,
    download_boards_and_sprints,
    download_fields,
    download_issuelinktypes,
    download_issues,
    download_issuetypes,
    download_priorities,
    download_projects_and_versions_and_components,
    download_resolutions,
    download_statuses,
    download_users,
    download_worklogs,
    get_ids_from_difference_of_issue_metadata,
    get_jira_search_batch_size,
)
from jf_ingest.utils import batch_iterable_by_bytes_size

logger = logging.getLogger(__name__)


class JiraObject(Enum):
    JiraFields = "jira_fields"
    JiraProjectsAndVersions = "jira_projects_and_versions"
    JiraUsers = "jira_users"
    JiraResolutions = "jira_resolutions"
    JiraIssueTypes = "jira_issuetypes"
    JiraLinkTypes = "jira_linktypes"
    JiraPriorities = "jira_priorities"
    JiraBoards = "jira_boards"
    JiraSprints = "jira_sprints"
    JiraBoardSprintLinks = "jira_board_sprint_links"
    JiraIssues = "jira_issues"
    JiraIssuesIdsDownloaded = "jira_issue_ids_downloaded"
    JiraIssuesIdsDeleted = "jira_issue_ids_deleted"
    JiraWorklogs = "jira_worklogs"
    JiraStatuses = "jira_statuses"


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_jira_to_s3(ingest_config: IngestionConfig):
    """Loads data from JIRA, Dumps it to disc, and then uploads that data to S3

    All configuration for this object is done via the JIRAIngestionConfig object

    Args:
        config (IngestionConfig): A dataclass that holds several different configuration args for this task

    Returns:
        bool: Returns True on Success
    """

    # For Agent, we leverage the DEBUG log level for some things that are really INFO level. We do this
    # because we want to keep the output console in Agent as clean as possible, and we don't want to
    # interrupt the TQDM progress bar
    #
    # When we run this code in Prefect, however, we lose the DEBUG level logs because Prefect logging is
    # pretty rigid and hard to set with fine granularity. To get around this problem, use logger.log with
    # a variable log level
    logging_helper._set_ingestion_type(ingest_config.ingest_type)
    logger.info('Beginning load_and_push_jira_to_s3')
    if not ingest_config.save_locally and not ingest_config.upload_to_s3:
        logger.error(
            f'Configuration error! Ingestion configuration must have either save_locally or upload_to_s3 set to True!'
        )
        raise Exception(
            'Save Locally and Upload to S3 are both set to False! Set one to true or no data will be saved!'
        )

    jira_config = ingest_config.jira_config
    if jira_config.skip_issues and jira_config.only_issues:
        logger.warning(f"only_issues and skip_issues are both True, so all tasks will be skipped. ")
        return False

    logging_helper.send_to_agent_log_file(f"Feature flags: {jira_config.feature_flags}")

    #######################################################################
    # SET UP JIRA CONNECTIONS (Basic and Potentially Atlassian Connect)
    #######################################################################
    jira_basic_connection = get_jira_connection(
        config=jira_config, auth_method=JiraAuthMethod.BasicAuth
    )

    jira_atlas_connect_connection = (
        get_jira_connection(config=jira_config, auth_method=JiraAuthMethod.AtlassianConnect)
        if JiraAuthMethod.AtlassianConnect in jira_config.available_auth_methods
        else None
    )
    # There is an ongoing effort to cut all things over to Atlassian Connect only,
    # but it is a piece wise migration for now.
    # OJ-29745
    jira_connect_or_fallback_connection = jira_basic_connection

    if jira_config.feature_flags.get("lusca-auth-always-use-connect-for-atlassian-apis-Q423"):
        logging_helper.send_to_agent_log_file("Will use connect for most API calls")
        jira_connect_or_fallback_connection = jira_atlas_connect_connection
    else:
        logging_helper.send_to_agent_log_file("Will use basic auth for most API calls")

    #######################################################################
    # Init IO Helper
    #######################################################################
    ingest_io_helper = IngestIOHelper(ingest_config=ingest_config)

    S3_OR_JELLYFISH_LOG_STATEMENT = (
        's3' if ingest_config.ingest_type == IngestionType.DIRECT_CONNECT else 'jellyfish'
    )
    if ingest_config.save_locally:
        logger.info(f"Data will be saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info("Data will not be saved locally")
    if ingest_config.upload_to_s3:
        logger.info(f"Data will be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data will NOT be submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    #######################################################################
    # Jira Projects
    #######################################################################
    projects_and_versions = download_projects_and_versions_and_components(
        jira_connection=jira_connect_or_fallback_connection,
        is_agent_run=ingest_config.ingest_type == IngestionType.AGENT,
        jellyfish_project_ids_to_keys=jira_config.jellyfish_project_ids_to_keys,
        jellyfish_issue_metadata=jira_config.jellyfish_issue_metadata,
        include_projects=jira_config.include_projects,
        exclude_projects=jira_config.exclude_projects,
        include_categories=jira_config.include_project_categories,
        exclude_categories=jira_config.exclude_project_categories,
    )

    project_ids = {proj["id"] for proj in projects_and_versions}
    ingest_io_helper.write_json_to_local_or_s3(
        object_name=JiraObject.JiraProjectsAndVersions.value,
        json_data=projects_and_versions,
        subdirectory=SubDirectory.JIRA,
        save_locally=ingest_config.save_locally,
        upload_to_s3=ingest_config.upload_to_s3,
    )

    if not jira_config.only_issues:
        #######################################################################
        # Jira Fields
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraFields.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_fields(
                jira_connect_or_fallback_connection,
                jira_config.include_fields,
                jira_config.exclude_fields,
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ######################################################################
        # Jira Users
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraUsers.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_users(
                jira_basic_connection=jira_basic_connection,  # Use BasicAuth because /users/search is not supported by Connect apps.
                jira_atlas_connect_connection=jira_atlas_connect_connection
                if jira_config.should_augment_emails
                else None,  # Use AtlasConnect for 'augment with email' subtask
                gdpr_active=jira_config.gdpr_active,
                search_users_by_letter_email_domain=jira_config.search_users_by_letter_email_domain,
                required_email_domains=jira_config.required_email_domains,
                is_email_required=jira_config.is_email_required,
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Resolutions
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraResolutions.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_resolutions(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Issue Types
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssueTypes.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_issuetypes(
                jira_connect_or_fallback_connection, project_ids=project_ids
            ),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Link Types
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraLinkTypes.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_issuelinktypes(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Priorities
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraPriorities.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_priorities(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Statuses
        #######################################################################
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraStatuses.value,
            subdirectory=SubDirectory.JIRA,
            json_data=download_statuses(jira_connect_or_fallback_connection),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Boards, Sprints, and Links
        #######################################################################
        boards, sprints, links = download_boards_and_sprints(
            jira_connect_or_fallback_connection,
            jira_config.download_sprints,
        )
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraBoards.value,
            subdirectory=SubDirectory.JIRA,
            json_data=boards,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraSprints.value,
            subdirectory=SubDirectory.JIRA,
            json_data=sprints,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraBoardSprintLinks.value,
            subdirectory=SubDirectory.JIRA,
            json_data=links,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

    if not jira_config.skip_issues:
        # For Jira Issue operations we need to determine the batch size that
        # the JIRA provider will limit us to
        jira_issues_batch_size = get_jira_search_batch_size(
            jira_connection=jira_connect_or_fallback_connection,
            optimistic_batch_size=jira_config.issue_batch_size,
        )
        logging_helper.send_to_agent_log_file(
            f"All issue operations will use a batch_size of {jira_issues_batch_size}",
        )

        #######################################################################
        # Pull Issue Metadata
        #######################################################################
        issue_metadata_from_jira: list[IssueMetadata] = download_all_issue_metadata(
            # We need jira 3.1.x to use connect: https://github.com/pycontribs/jira/pull/1157
            # So for now force basic auth.
            jira_connection=jira_basic_connection,
            project_keys=[
                proj["key"]
                for proj in projects_and_versions
                if proj["key"] not in jira_config.exclude_projects
            ],
            earliest_issue_dt=jira_config.earliest_issue_dt,
            num_parallel_threads=jira_config.issue_download_concurrent_threads,
            batch_size=jira_issues_batch_size,
        )

        issues_generator = download_issues(
            jira_connection=jira_connect_or_fallback_connection,
            full_redownload=jira_config.full_redownload,
            issue_download_concurrent_threads=jira_config.issue_download_concurrent_threads,
            jira_issues_batch_size=jira_issues_batch_size,
            issue_metadata_from_jellyfish=jira_config.jellyfish_issue_metadata,
            issue_metadata_from_jira=issue_metadata_from_jira,
            include_fields=jira_config.include_fields,
            exclude_fields=jira_config.exclude_fields,
        )

        jira_issues_uncompressed_file_size = 50 * Constants.MB_SIZE_IN_BYTES
        all_downloaded_issue_ids = []
        total_issue_batches = 0

        for batch_number, batch_issues in enumerate(
            batch_iterable_by_bytes_size(
                issues_generator, batch_byte_size=jira_issues_uncompressed_file_size
            )
        ):
            logging_helper.send_to_agent_log_file(
                f'Saving {len(batch_issues)} issues as batch number {batch_number}',
            )
            all_downloaded_issue_ids.extend([issue['id'] for issue in batch_issues])
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraIssues.value,
                subdirectory=SubDirectory.JIRA,
                json_data=batch_issues,
                batch_number=batch_number,
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )
            total_issue_batches += 1

        logger.info(
            f"Successfully saved {len(all_downloaded_issue_ids)} Jira Issues in {total_issue_batches} separate batches, with each batch limited to {round(jira_issues_uncompressed_file_size / Constants.MB_SIZE_IN_BYTES)}MB per batch"
        )

        # Write Issues
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssuesIdsDownloaded.value,
            json_data=[int(issue_id) for issue_id in all_downloaded_issue_ids],
            subdirectory=SubDirectory.JIRA,
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        deleted_issue_ids = get_ids_from_difference_of_issue_metadata(
            jira_config.jellyfish_issue_metadata, issue_metadata_from_jira
        )

        logger.info(f'{len(deleted_issue_ids)} issues have been detected as being deleted')
        # Write issues that got deleted
        ingest_io_helper.write_json_to_local_or_s3(
            object_name=JiraObject.JiraIssuesIdsDeleted.value,
            subdirectory=SubDirectory.JIRA,
            json_data=list(deleted_issue_ids),
            save_locally=ingest_config.save_locally,
            upload_to_s3=ingest_config.upload_to_s3,
        )

        #######################################################################
        # Jira Work Logs
        #######################################################################
        if jira_config.download_worklogs:
            ingest_io_helper.write_json_to_local_or_s3(
                object_name=JiraObject.JiraWorklogs.value,
                subdirectory=SubDirectory.JIRA,
                json_data=download_worklogs(
                    jira_connect_or_fallback_connection,
                    all_downloaded_issue_ids,
                    jira_config.work_logs_pull_from,
                ),
                save_locally=ingest_config.save_locally,
                upload_to_s3=ingest_config.upload_to_s3,
            )
    else:
        logger.info(
            f"Skipping issues and worklogs bc config.skip_issues is {jira_config.skip_issues}"
        )

    if ingest_config.save_locally:
        logger.info(f"Data has been saved locally to: {ingest_io_helper.local_file_path}")
    else:
        logger.info(
            f"Data has not been saved locally, because save_locally was set to false in the ingest config!"
        )

    if ingest_config.upload_to_s3:
        logger.info(f"Data has been submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")
    else:
        logger.info(f"Data was not submitted to {S3_OR_JELLYFISH_LOG_STATEMENT}")

    return True



import json
import logging
import time
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Generator

import requests
from requests import Response, Session

from jf_ingest.jf_git.exceptions import GqlRateLimitExceededException
from jf_ingest.utils import retry_for_429s

logger = logging.getLogger(__name__)


GQL_PAGE_INFO_BLOCK = "pageInfo {hasNextPage, endCursor}"
GQL_RATE_LIMIT_QUERY_BLOCK = "rateLimit {remaining, resetAt}"


@dataclass
class GQLRateLimit:
    remaining: int
    reset_at: datetime


class GqlRateLimitedExceptionInner(Exception):
    pass


def gql_format_to_datetime(datetime_str: str) -> datetime:
    """Attempt to formate a datetime str from the GQL format to a python Datetime Object
    NOTE: This currently is only verified to support the github GQL format. It is NOT YET
    GENERALIZED

    Args:
        datetime_str (str): The datetime from graphql

    Returns:
        datetime: A valid, timezone aware datetime
    """
    return datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)


def get_gql_rate_limit(session: Session, base_url: str) -> GQLRateLimit:
    """Attempt to pull the current rate limit information from GQL
    NOTE: Getting the rate limit info is never affected by the current rate limit

    Args:
        session (Session): A valid session connecting us to the GQL API
        base_url (str): The base URL we are hitting

    Returns:
        dict: A dictionary object containing rate limit information (remaing and resetAt)
    """
    query_body = f"{{{GQL_RATE_LIMIT_QUERY_BLOCK}}}"
    # NOTE: DO NOT CALL get_raw_gql_result TO GET THE RESULTS HERE! IT'S A RECURSIVE TRAP
    response: Response = retry_for_429s(session.post, url=base_url, json={'query': query_body})
    response.raise_for_status()
    json_str = response.content.decode()
    raw_data: dict = json.loads(json_str)['data']['rateLimit']
    return GQLRateLimit(
        remaining=int(raw_data['remaining']), reset_at=gql_format_to_datetime(raw_data['resetAt'])
    )


def get_raw_gql_result(
    session: Session, gql_base_url: str, query_body: str, max_attempts: int = 7
) -> dict:
    """Gets the raw results from a Graphql Query.

    Args:
        session (Session): A valid session object connecting us to the GQL API
        gql_base_url (str): The base URL for the GQL API
        query_body (str): A query body to hit GQL with
        max_attempts (int, optional): The number of retries we should make when we specifically run into GQL Rate limiting. This value is important if the GQL endpoint doesn't give us (or gives us a malformed) rate limit header. Defaults to 7.

    Raises:
        GqlRateLimitExceededException: A custom exception if we run into GQL rate limiting and we run out of attempts (based on max_attempts)
        Exception: Any other random exception we encounter, although the big rate limiting use cases are generally covered

    Returns:
        dict: A raw dictionary result from GQL
    """
    attempt_number = 1
    while True:
        try:
            response: Response = retry_for_429s(
                session.post, url=gql_base_url, json={'query': query_body}
            )
            response.raise_for_status()
            json_str = response.content.decode()
            json_data = json.loads(json_str)
            if 'errors' in json_data:
                if len(json_data['errors']) == 1:
                    error_dict: dict = json_data['errors'][0]
                    if error_dict.get('type') == 'RATE_LIMITED':
                        raise GqlRateLimitedExceptionInner(
                            error_dict.get('message', 'Rate Limit hit in GQL')
                        )
                raise Exception(
                    f'Exception encountered when trying to query: {query_body}. Error: {json_data["errors"]}'
                )
            return json_data
        # We can get transient 403 level errors that have to do with rate limiting,
        # but aren't directly related to the above GqlRateLimitedExceptionInner logic.
        # Do a simple retry loop here
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 403:
                raise
            if attempt_number > max_attempts:
                raise

            sleep_time = attempt_number**2
            # Overwrite sleep time if github gives us a specific wait time
            if e.response.headers.get('retry-after') and attempt_number == 1:
                retry_after = int(e.response.headers.get('retry-after'))
                if retry_after > (60 * 5):
                    # if the given wait time is more than 5 minutes, call their bluff
                    # and try the experimental backoff approach
                    pass
                elif retry_after <= 0:
                    # if the given wait time is negative ignore their suggestion
                    pass
                else:
                    # Add three seconds for gracetime
                    sleep_time = retry_after + 3

            logger.warning(
                f'A secondary rate limit was hit. Sleeping for {sleep_time} seconds. (attempt {attempt_number}/{max_attempts})',
            )
            time.sleep(sleep_time)
        except GqlRateLimitedExceptionInner:
            if attempt_number > max_attempts:
                raise GqlRateLimitExceededException(
                    f'Exceeded maximum retry limit ({max_attempts})'
                )

            rate_limit_info: GQLRateLimit = get_gql_rate_limit(
                session=session, base_url=gql_base_url
            )
            reset_at: datetime = rate_limit_info.reset_at
            reset_at_timestamp = reset_at.timestamp()
            curr_timestamp = datetime.utcnow().timestamp()

            sleep_time = reset_at_timestamp - curr_timestamp

            # Sometimes github gives a reset time way in the
            # future. But rate limits reset each hour, so don't
            # wait longer than that
            sleep_time = min(sleep_time, 3600)

            # Sometimes github gives a reset time in the
            # past. In that case, wait for 5 mins just in case.
            if sleep_time <= 0:
                sleep_time = 300
            logger.warning(
                f'GQL Rate Limit hit. Sleeping for {sleep_time} seconds',
            )
            time.sleep(sleep_time)
        finally:
            attempt_number += 1


def page_results_gql(
    session: Session,
    gql_base_url: str,
    query_body: str,
    path_to_page_info: str,
    cursor: str = 'null',
) -> Generator[dict, None, None]:
    """This is a helper function for paging results from GraphQL. It expects
    a query body to hit Graphql with that has a %s marker after the "after:"
    key word, so that we can inject a cursor into the query. This will allow
    us to page results in GraphQL.
    To use this function properly, the section you are trying to page MUST
    INCLUDE VALID PAGE INFO (including the hasNext and endCursor attributes)

    Args:
        query_body (str): The query body to hit GraphQL with
        path_to_page_info (str): A string of period separated words that lead
        to the part of the query that we are trying to page. Example: data.organization.userQuery
        cursor (str, optional): LEAVE AS NULL - this argument is use recursively to page. The cursor
        will continuously go up, based on the endCursor attribute in the GQL call. Defaults to 'null'.

    Yields:
        Generator[dict, None, None]: This function yields each item from all the pages paged, item by item
    """
    hasNextPage = True
    while hasNextPage:
        # Fetch results
        result = get_raw_gql_result(
            session=session, gql_base_url=gql_base_url, query_body=(query_body % cursor)
        )

        yield result

        # Get relevant data and yield it
        path_tokens = path_to_page_info.split('.')
        for token in path_tokens:
            result = result[token]

        page_info = result['pageInfo']
        # Need to grab the cursor and wrap it in quotes
        _cursor = page_info['endCursor']
        # If endCursor returns null (None), break out of loop
        hasNextPage = page_info['hasNextPage'] and _cursor
        cursor = f'"{_cursor}"'



import logging
import traceback
from datetime import datetime
from typing import Generator, List

from dateutil import parser
from requests.exceptions import HTTPError

from jf_ingest import diagnostics, logging_helper
from jf_ingest.config import GitConfig
from jf_ingest.graphql_utils import gql_format_to_datetime
from jf_ingest.jf_git.adapters import (
    GitAdapter,
    StandardizedBranch,
    StandardizedCommit,
    StandardizedOrganization,
    StandardizedPullRequest,
    StandardizedPullRequestComment,
    StandardizedPullRequestReview,
    StandardizedRepository,
    StandardizedShortRepository,
    StandardizedTeam,
    StandardizedUser,
)
from jf_ingest.jf_git.clients.github import GithubClient
from jf_ingest.name_redactor import NameRedactor, sanitize_text

logger = logging.getLogger(__name__)

'''

    Data Fetching

'''

_branch_redactor = NameRedactor(preserve_names=['master', 'develop'])
_organization_redactor = NameRedactor()
_repo_redactor = NameRedactor()


class GithubAdapter(GitAdapter):
    def __init__(self, config: GitConfig):
        # Git Config options
        self.client = GithubClient(config.git_auth_config)
        self.config = config
        self.repo_id_to_name_lookup: dict = {}
        self.repo_to_branch_is_quiescent_lookups: dict = {}
        self.repo_has_quiescent_prs_lookup: dict = {}

    def get_api_scopes(self):
        return self.client.get_scopes_of_api_token()

    @diagnostics.capture_timing()
    @logging_helper.log_entry_exit(logger)
    def get_organizations(self) -> List[StandardizedOrganization]:
        # NOTE: For github, organization equates to a Github organization here!
        # We have the name of the organization in the config, expected to be a list of length one
        try: 
            org = self.client.get_organization_by_login(self.config.git_organizations[0])
        except HTTPError as e:
            if e.response.status_code == 404:
                # give something a little nicer for 404s
                raise ValueError(
                    'Organization not found. Make sure your token has appropriate access to Github.'
                )
            raise
        return [
            _standardize_organization(org, self.config.git_redact_names_and_urls)
        ]

    def get_users(
        self, standardized_organization: StandardizedOrganization, limit: int = None
    ) -> Generator[StandardizedUser, None, None]:
        for i, user in enumerate(self.client.get_users(standardized_organization.login), start=1):
            yield _standardize_user(user)
            if limit and i >= limit:
                return

    def get_teams(
        self, standardized_organization: StandardizedOrganization, limit: int = None
    ) -> Generator[StandardizedTeam, None, None]:
        for i, team in enumerate(
            self.client.get_teams(login=standardized_organization.login), start=1
        ):
            yield _standardize_team(team)
            if limit and i >= limit:
                return

    def get_repos(
        self, standardized_organization: StandardizedOrganization,
    ) -> Generator[StandardizedRepository, None, None]:
        filters = []
        if self.config.excluded_repos:
            filters.append(
                lambda r: r['name'].lower() in set([r.lower() for r in self.config.included_repos])
            )
        if self.config.excluded_repos:
            filters.append(
                lambda r: r['name'].lower()
                not in set([r.lower() for r in self.config.excluded_repos])
            )

        repo_found = False
        for api_repo in self.client.get_repos(
            standardized_organization.login, repo_filters=filters
        ):
            repo_found = True
            # Enter the repo to the ID to Name look up, incase the repo name gets
            # scrubbed by our git_redact_names_and_urls logic
            repo_id = api_repo['id']
            self.repo_id_to_name_lookup[repo_id] = api_repo['name']

            # Mark if the default branch in this repo is quiescent, to save on API calls later
            # Initiate branch lookup table
            self.repo_to_branch_is_quiescent_lookups[repo_id] = {}
            default_branch = api_repo.get('defaultBranch')
            if default_branch:
                commits = (default_branch['target']['mostRecentCommits'] or {}).get('commits', [])
                if commits:
                    # Translate data from API
                    most_recent_commit = commits[0]
                    most_recent_commit_date: datetime = gql_format_to_datetime(
                        most_recent_commit['committedDate']
                    )
                    # Get our internal 'pull_from' value
                    pull_since_for_commits = self.config.get_pull_from_for_commits(
                        org_login=standardized_organization.login, repo_id=repo_id
                    )

                    # Mark if we can skip pulling this branch or not
                    self.repo_to_branch_is_quiescent_lookups[repo_id][default_branch['name']] = (
                        pull_since_for_commits >= most_recent_commit_date
                    )
                else:
                    # If there are no commits, there is nothing to pull
                    # Mark as quiescent
                    self.repo_to_branch_is_quiescent_lookups[repo_id][default_branch['name']] = True

            # Mark the latest date for PRs in this repo, to save on API calls later
            prs = api_repo['prQuery'].get('prs', [])
            if prs:
                # Translate latest PR detected from API
                latest_pr_update = gql_format_to_datetime(prs[0]['updatedAt'])
                # Get our own pull since value
                pull_since_for_prs = self.config.get_pull_from_for_prs(
                    org_login=standardized_organization.login, repo_id=repo_id
                )

                # Mark if we can skip this PR or not
                self.repo_has_quiescent_prs_lookup[repo_id] = pull_since_for_prs >= latest_pr_update
            else:
                self.repo_has_quiescent_prs_lookup[repo_id] = True

            yield _standardize_repo(
                api_repo, standardized_organization, self.config.git_redact_names_and_urls
            )

        if not repo_found:
            raise ValueError(
                'No repos found. Make sure your token has appropriate access to Github and check your configuration of repos to pull.'
            )

    def get_commits_for_included_branches(
        self, standardized_repo: StandardizedRepository,
    ) -> Generator[StandardizedCommit, None, None]:
        pull_since = self.config.get_pull_from_for_commits(
            org_login=standardized_repo.organization.login, repo_id=standardized_repo.id
        )

        for branch_name in self.get_branches_for_standardized_repo(standardized_repo):
            if self.repo_to_branch_is_quiescent_lookups[standardized_repo.id].get(
                branch_name, False
            ):
                continue
            try:
                login = standardized_repo.organization.login
                repo_name = self.repo_id_to_name_lookup[standardized_repo.id]
                for j, api_commit in enumerate(
                    self.client.get_commits(
                        login=login, repo_name=repo_name, branch_name=branch_name, since=pull_since,
                    ),
                    start=1,
                ):
                    with logging_helper.log_loop_iters('branch commit inside repo', j, 100):
                        yield _standardize_commit(
                            api_commit,
                            standardized_repo,
                            branch_name,
                            self.config.git_strip_text_content,
                            self.config.git_redact_names_and_urls,
                        )

            except Exception as e:
                logging_helper.send_to_agent_log_file(traceback.format_exc())
                logger.warning(f':WARN: Got exception for branch {branch_name}: {e}. Skipping...')

    def get_commits_for_default_branch(
        self, standardized_repo: StandardizedRepository, limit: int = None
    ) -> Generator[StandardizedCommit, None, None]:
        try:
            pull_since = self.config.get_pull_from_for_commits(
                org_login=standardized_repo.organization.login, repo_id=standardized_repo.id
            )
            login = standardized_repo.organization.login
            repo_name = self.repo_id_to_name_lookup[standardized_repo.id]
            branch_name = standardized_repo.default_branch_name
            for j, api_commit in enumerate(
                self.client.get_commits(
                    login=login, repo_name=repo_name, branch_name=branch_name, since=pull_since,
                ),
                start=1,
            ):
                with logging_helper.log_loop_iters('branch commit inside repo', j, 100):
                    yield _standardize_commit(
                        api_commit,
                        standardized_repo,
                        standardized_repo.default_branch_name,
                        self.config.git_strip_text_content,
                        self.config.git_redact_names_and_urls,
                    )
                    if limit and j >= limit:
                        return

        except Exception as e:
            logging_helper.send_to_agent_log_file(traceback.format_exc())
            logger.warning(
                f':WARN: Got exception for branch {standardized_repo.default_branch_name}: {e}. Skipping...'
            )

    def get_prs(
        self, standardized_repo: StandardizedRepository, limit: int = None
    ) -> Generator[StandardizedPullRequest, None, None]:
        try:
            login = standardized_repo.organization.login
            repo_id = standardized_repo.id
            repo_name = self.repo_id_to_name_lookup[repo_id]
            # Check if we flagged this repo as quiescent
            if self.repo_has_quiescent_prs_lookup[repo_id]:
                return

            pull_since = self.config.get_pull_from_for_prs(login, repo_id)

            api_prs = self.client.get_prs(
                login=login,
                repo_name=repo_name,
                include_top_level_comments=self.config.jf_options.get(
                    'get_all_issue_comments', False
                ),
            )
            for j, api_pr in enumerate(api_prs, start=1,):
                with logging_helper.log_loop_iters('pr inside repo', j, 10):
                    try:
                        updated_at = parser.parse(api_pr['updatedAt'])

                        # PRs are ordered newest to oldest
                        # if this is too old, we're done with this repo.
                        # This is an INCLUSIVE check, to stop us from grabbing
                        # the last PR that is our marker for pull_since!
                        if pull_since and pull_since >= updated_at:
                            break

                        yield _standardize_pr(
                            api_pr,
                            standardized_repo,
                            self.config.git_strip_text_content,
                            self.config.git_redact_names_and_urls,
                        )
                        if limit and j >= limit:
                            return
                    except Exception:
                        # if something goes wrong with normalizing one of the prs - don't stop pulling. try
                        # the next one.
                        pr_id = f' {api_pr["id"]}' if api_pr else ''
                        logger.warning(
                            f'normalizing PR {pr_id} from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...'
                        )
                        logging_helper.send_to_agent_log_file(traceback.format_exc())

        except Exception:
            # if something happens when pulling PRs for a repo, just keep going.
            logger.warning(
                f'normalizing PRs from repo {standardized_repo.name} ({standardized_repo.id}). Skipping...'
            )
            logging_helper.send_to_agent_log_file(traceback.format_exc())


'''

    Massage Functions

'''


def _standardize_user(api_user) -> StandardizedUser:
    if not api_user:
        return None

    id = api_user.get('id')
    name = api_user.get('name')
    login = api_user.get('login')
    email = api_user.get('email')
    # raw user, just have email (e.g. from a commit)
    if not id:
        return StandardizedUser(id=email, login=email, name=name, email=email,)

    # API user, where github matched to a known account
    return StandardizedUser(id=id, login=login, name=name, email=email)


def _standardize_team(api_team: dict) -> StandardizedTeam:
    return StandardizedTeam(
        id=str(api_team.get('id')),
        name=api_team.get('name'),
        slug=api_team.get('slug'),
        description=api_team.get('description'),
        members=[_standardize_user(member) for member in api_team.get('members', [])],
    )


def _standardize_organization(
    api_org: dict, redact_names_and_urls: bool
) -> StandardizedOrganization:
    return StandardizedOrganization(
        id=api_org['id'],
        login=api_org['login'],
        name=api_org.get('name')
        if not redact_names_and_urls
        else _organization_redactor.redact_name(api_org.get('name')),
        url=api_org['url'] if not redact_names_and_urls else None,
    )


def _standardize_branch(api_branch, redact_names_and_urls: bool) -> StandardizedBranch:
    return StandardizedBranch(
        name=api_branch['name']
        if not redact_names_and_urls
        else _branch_redactor.redact_name(api_branch['name']),
        sha=api_branch['target']['sha'],
    )


def _standardize_repo(
    api_repo, standardized_organization: StandardizedOrganization, redact_names_and_urls: bool,
) -> StandardizedRepository:
    repo_name = (
        api_repo['name']
        if not redact_names_and_urls
        else _repo_redactor.redact_name(api_repo['name'])
    )
    url = api_repo['url'] if not redact_names_and_urls else None

    standardized_branches = [
        _standardize_branch(branch, redact_names_and_urls) for branch in api_repo['branches']
    ]

    # If a repsotiroy is completely empty, than there will be no default branch
    # this can mess with our ingestion due to the 'name' field being nested
    # on a None object
    default_branch_name = api_repo['defaultBranch']['name'] if api_repo['defaultBranch'] else None
    return StandardizedRepository(
        id=api_repo['id'],
        name=repo_name,
        full_name=repo_name,
        url=url,
        default_branch_name=default_branch_name,
        is_fork=api_repo['isFork'],
        branches=standardized_branches,
        organization=standardized_organization,
    )


def _standardize_short_form_repo(
    api_repo: dict, redact_names_and_urls: dict
) -> StandardizedShortRepository:
    repo_name = (
        api_repo['name']
        if not redact_names_and_urls
        else _repo_redactor.redact_name(api_repo['name'])
    )
    url = api_repo['url'] if not redact_names_and_urls else None

    return StandardizedShortRepository(id=api_repo['id'], name=repo_name, url=url)


def _standardize_commit(
    api_commit: dict,
    standardized_repo: StandardizedRepository,
    branch_name: str,
    strip_text_content: bool,
    redact_names_and_urls: bool,
):
    author = _standardize_user(api_commit['author'])
    commit_url = api_commit['url'] if not redact_names_and_urls else None
    return StandardizedCommit(
        hash=api_commit['sha'],
        author=author,
        url=commit_url,
        commit_date=api_commit['committedDate'],
        author_date=api_commit['authoredDate'],
        message=sanitize_text(api_commit['message'], strip_text_content),
        is_merge=api_commit['parents']['totalCount'] > 1,
        repo=standardized_repo.short(),  # use short form of repo
        branch_name=branch_name
        if not redact_names_and_urls
        else _branch_redactor.redact_name(branch_name),
    )


def _get_standardized_pr_comments(
    api_comments: list[dict], strip_text_content
) -> List[StandardizedPullRequestComment]:
    return [
        StandardizedPullRequestComment(
            user=_standardize_user(api_comment['author']),
            body=sanitize_text(api_comment['body'], strip_text_content),
            created_at=api_comment['createdAt'],
        )
        for api_comment in api_comments
    ]


def _get_standardized_reviews(api_reviews: list[dict]):
    return [
        StandardizedPullRequestReview(
            user=_standardize_user(api_review['author']),
            foreign_id=api_review['id'],
            review_state=api_review['state'],
        )
        for api_review in api_reviews
    ]


def _standardize_pr(
    api_pr,
    standardized_repo: StandardizedRepository,
    strip_text_content: bool,
    redact_names_and_urls: bool,
):
    base_branch_name = api_pr['baseRefName']
    head_branch_name = api_pr['headRefName']
    standardized_merge_commit = (
        _standardize_commit(
            api_pr['mergeCommit'],
            standardized_repo=standardized_repo,
            branch_name=base_branch_name,
            strip_text_content=strip_text_content,
            redact_names_and_urls=redact_names_and_urls,
        )
        if api_pr['mergeCommit']
        else None
    )
    return StandardizedPullRequest(
        id=api_pr['id'],
        additions=api_pr['additions'],
        deletions=api_pr['deletions'],
        changed_files=api_pr['changedFiles'],
        created_at=api_pr['createdAt'],
        updated_at=api_pr['updatedAt'],
        merge_date=api_pr['mergedAt'],
        closed_date=api_pr['closedAt'],
        is_closed=api_pr['state'].lower() == 'closed',
        is_merged=api_pr['merged'],
        # redacted fields
        url=api_pr['url'] if not redact_names_and_urls else None,
        base_branch=(
            base_branch_name
            if not redact_names_and_urls
            else _branch_redactor.redact_name(base_branch_name)
        ),
        head_branch=(
            head_branch_name
            if not redact_names_and_urls
            else _branch_redactor.redact_name(head_branch_name)
        ),
        # sanitized fields
        title=sanitize_text(api_pr['title'], strip_text_content),
        body=sanitize_text(api_pr['body'], strip_text_content),
        # standardized fields
        commits=[
            _standardize_commit(
                api_commit=commit,
                standardized_repo=standardized_repo,
                branch_name=base_branch_name,
                strip_text_content=strip_text_content,
                redact_names_and_urls=redact_names_and_urls,
            )
            for commit in api_pr['commits']
        ],
        merge_commit=standardized_merge_commit,
        author=_standardize_user(api_user=api_pr['author']),
        merged_by=_standardize_user(api_user=api_pr['mergedBy']),
        approvals=_get_standardized_reviews(api_pr['reviews']),
        comments=_get_standardized_pr_comments(api_pr['comments'], strip_text_content),
        base_repo=_standardize_short_form_repo(api_pr['baseRepository'], redact_names_and_urls),
        head_repo=_standardize_short_form_repo(api_pr['baseRepository'], redact_names_and_urls),
    )
